import { describe, test, beforeEach } from 'node:test';
import assert from 'node:assert/strict';

class Item {
  constructor(name, price, category) {
    this.name = name;
    this.price = price;
    this.category = category;
  }
}

class SimplifiedInvoice {
  constructor({confirmationNumber, paymentMethod, paymentDate}) {
    this.confirmationNumber = confirmationNumber;
    this.paymentMethod = paymentMethod;
    this.paymentDate = paymentDate;
    this.items = new Map();
  }
  add(item, amount) {
    this.items.set(item, amount);
  }
  total() {
    let eightTotal = 0;
    let tenTotal = 0;
    for (const [item, amount] of this.items) {
      if(item.category.tax_type === 'reduced') {
        eightTotal += (item.price * amount);
      } else {
        tenTotal += (item.price * amount);
      }
    }
    const eightTaxTotal = Math.floor(eightTotal * (8/108));
    const tenTaxTotal = Math.floor(tenTotal * (10/110));
    return {
      total: eightTotal + tenTotal,
      sub_total: {
        reduced: eightTotal,
        standard: tenTotal
      },
      tax_included: true,
      tax: {
        total: eightTaxTotal + tenTaxTotal,
        reduced: eightTaxTotal,
        standard: tenTaxTotal
      }
    };
  }
}

describe('情報が少なすぎる', () => {
  let inv;
  let beverage;
  let liquor;

  beforeEach(() => {
    inv = new SimplifiedInvoice({
      confirmationNumber: 'XXXX-XXXXX-XXXXX',
      paymentDate: new Date()
    });
    beverage = {
      type: 'beverage',
      name: '飲料',
      tax_type: 'reduced'
    };
    liquor = {
      type: 'liquor',
      name: '酒類',
      tax_type: 'standard'
    };
  });

  function addTea(n) {
    const item = new Item('技評茶', 130, beverage);
    inv.add(item, n);
  }

  function addBeer(n) {
    const beer = new Item('技評ビール', 150, liquor);
    inv.add(beer, n);
  }

  test('税額計算結果が正しいこと', () => {
    addTea(2);
    addBeer(3);
    const { reduced, standard } = inv.total().tax;
    assert.equal(reduced, 19);
    assert.equal(standard, 40);
  });
});

describe('情報が多すぎる', () => {
  test('税額計算結果が正しいこと', () => {
    const inv = new SimplifiedInvoice({
      confirmationNumber: 'XXXX-XXXXX-XXXXX',
      paymentMethod: 'cash',
      paymentDate: new Date()
    });
    inv.add(new Item('技評茶', 130, {
      type: 'beverage',
      name: '飲料',
      tax_type: 'reduced'
    }), 2);
    inv.add(new Item('技評酒', 150, {
      type: 'liquor',
      name: '酒類',
      tax_type: 'standard'
    }), 3);
    const res = inv.total();
    assert.equal(res.total, 710);
    assert.equal(res.sub_total.reduced, 260);
    assert.equal(res.sub_total.standard, 450);
    assert.equal(res.tax_included, true);
    assert.equal(res.tax.total, 59);
    assert.equal(res.tax.reduced, 19);
    assert.equal(res.tax.standard, 40);
  });
});

describe('適格簡易請求書における消費税額等の計算と端数処理', () => {
  const 飲料 = {
    type: 'beverage',
    name: '飲料',
    tax_type: 'reduced'
  };
  const 酒類 = {
    type: 'liquor',
    name: '酒類',
    tax_type: 'standard'
  };
  function createSimplifiedInvoice () {
    return new SimplifiedInvoice({
      confirmationNumber: 'XXXX-XXXXX-XXXXX',
      paymentMethod: 'cash',
      paymentDate: new Date()
    });
  }

  describe('適格簡易請求書の消費税額計算', () => {
    test('税込価額を税率ごとに区分して合計した金額に対して税額を計算し、端数は切り捨てること', () => {
      // 準備（Arrange, Given）
      // 軽減税率の商品（飲料）と標準税率の商品（酒類）を
      // それぞれ複数購入している事前状態において
      const inv = createSimplifiedInvoice();
      inv.add(new Item('技評茶', 130, 飲料), 2);
      inv.add(new Item('技評酒', 150, 酒類), 3);

      // 実行（Act, When）
      // 合計、小計、消費税額などの計算を行った場合
      const total = inv.total();

      // 検証（Assert, Then）
      // 税率ごとの合計に対して1回ずつ切り捨てられること
      assert.deepEqual(total.tax, {
        reduced: 19,  // (130*2)*(8/108)  => 19.25円
        standard: 40, // (150*3)*(10/110) => 40.90円
        total: 59     // 消費税額は合算するだけ
      });
    });
  });
});
